# -*- coding: utf-8; indent-tabs-mode: nil -*-

# mailbridge: a mail storage spanning bridge
# 
# Copyright 2011 Ryan Hughes <ryan@n-h.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import binascii

import MAPI
from MAPI.Util import *


class KeyError(Exception):
    def __init__(self): pass

class RedisStateStreamStorage:
    def __init__(self):
        self.basekey = None
        self.store = None
        
    def bind_store(self, server, server_store, store):
        self.store = store
        self.basekey = self.store.get_redis_basekey()
        keyname = '%s:server_store' % (self.basekey)
        server_store_id = "%s:%s" % (binascii.hexlify(server.server_guid), binascii.hexlify(server_store.store_entryid))
        last_server_store_id = self.store.redis.get(keyname)
        if (last_server_store_id == None):
            self.store.redis.set(keyname, server_store_id)
            return True
        else:
            if (last_server_store_id == server_store_id):
                return True
            else:
                return False
        
    def get_redis_basekey(self):
        return self.basekey

    def load_state(self, entryid):
        keyname = '%s:state' % (self.basekey)
        stream = None
        try:
            hexdata = self.store.redis.hget(keyname, binascii.hexlify(entryid))
            if hexdata == None:
                return stream
            bin = binascii.unhexlify(hexdata)
            stream = IStream()
            stream.Write(bin)
        except IOError:
            pass
        return stream

    def save_state(self, entryid, stream):
        keyname = '%s:state' % (self.basekey)
        clone = stream.Clone()
        clone.Seek(BOOKMARK_BEGINNING, 0)
        bin = clone.Read(100000000) #0xffffffff    #TODO why does this faile
        self.store.redis.hset(keyname, binascii.hexlify(entryid), binascii.hexlify(bin))

        