# -*- coding: utf-8; indent-tabs-mode: nil -*-

# mailbridge: a mail storage spanning bridge
# 
# Copyright 2011 Ryan Hughes <ryan@n-h.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from mailbridge.store.BaseMailbox import BaseMailbox
from mailbridge.store.Message import Message
from mailbridge.store.Change import ReadState
import logging

CR = '\r'
LF = '\n'
CRLF = CR+LF


class DsyncMailbox(BaseMailbox):
    '''
    classdocs
    '''


    def __init__(self, store, guid):
        '''
        Constructor
        '''
        super(DsyncMailbox, self).__init__(store, guid)
        self.Changes = None
        self.selected = False
        self.getMessageList()
        self.last_modseq = store.get_folder_max_modseq(guid)
        self.uid_next = store.get_folder_max_uid_next(guid)
        self.first_sequence_id = self.uid_next
        self.logger = logging.getLogger("DsyncMailbox")
        
    def nextuid(self):
        uid_next = self.uid_next
        self.uid_next += 1
        return uid_next

    def getMessageList(self):
        # process added messages and combine with uids
        self.store.connection.boxselect(self.guid)
        msg_list_data_lines = self.store.connection.msglist([self.guid])[0]
        msg_list_data = {}
        for msg_list_item_line in msg_list_data_lines:
            # box_index, msg_guid, uid, modseq, flags, save_date = msg_list_item_line.split('\t')
            msg_list_data_item = msg_list_item_line.split('\t')
            msg_guid = msg_list_data_item[1]
            msg_uid = msg_list_data_item[2]
            if (not msg_list_data.has_key(msg_guid)):
                msg_list_data[msg_guid] = msg_list_data_item
        self.msg_list_data = msg_list_data
        
    def getMessageListForMinModseq(self, min_modseq):
        # process added messages and combine with uids
        self.store.connection.boxselect(self.guid)
        msg_list_data_lines = self.store.connection.msglist([self.guid])[0]
        msg_list_data = []
        for msg_list_item_line in msg_list_data_lines:
            # box_index, msg_guid, uid, modseq, flags, save_date = msg_list_item_line.split('\t')
            if (msg_list_item_line[0] == '+'):
                break
            msg_list_data_item = msg_list_item_line.split('\t')
            msg_modseq = int(msg_list_data_item[3])
            if msg_modseq < min_modseq: continue
            msg_uid = int(msg_list_data_item[2])
            msg_list_data.append(msg_list_data_item)
        return msg_list_data
        
    def get_redis_basekey(self, replica_id=None):
        store_basekey = self.store.get_redis_basekey(replica_id=replica_id)
        return "%s:%s" % (store_basekey, self.guid)
    
    def getUidForGuid(self, id):
        uidstr = self.store.redis.hget(self.get_redis_basekey()+":sourceid2uid",id)
        if uidstr != None:
            return int(uidstr)
        else:
            return None
        
        if self.msg_list_data.has_key(id):
            uid = self.msg_list_data[id][2]
            return int(uid)
        else:
            return None

    def getMessageInfo(self, index):
        pass

    def addMessage(self, id, message):
        # track added msg guids after max modseq
        if not self.selected:
            self.store.connection.boxselect(self.guid)
            self.selected = True
        flags = self.getFlagsFromMessage(message)
        sequence_id = self.nextuid()
        self.logger.debug("LASTUID: '%s' " % sequence_id)
        self.store.redis.hset(self.get_redis_basekey()+":sequenceid2id",sequence_id,id)
        self.store.redis.hset(self.get_redis_basekey()+":sequenceid2flags",sequence_id,flags)
        self.store.connection.msgsave(message.received_date, '1', id, sequence_id, '0', flags, 
                  message.received_date,
                  message.rfc822lines())
        self.store.redis.rpush(self.get_redis_basekey()+":transactionlog",sequence_id)

    def updateMessage(self, id, message):
        last_message = Message()
        uid = self.getUidForGuid(id)
        if uid:
            last_flags = self.store.redis.hget(self.get_redis_basekey()+":uid2flags",uid)
            if last_flags != None:
                self.getMessageFlagsFromFlags(last_flags, last_message)
        message.extra_flags = last_message.extra_flags
        self.delMessages([id])
        self.addMessage(id, message)

    def updateMessageFlags(self, id, flags):
        if not self.selected:
            self.store.connection.boxselect(self.guid)
            self.selected = True
        uid = self.getUidForGuid(id)
        if uid:
            self.store.connection.msgupdate(uid, 0, flags)
            self.store.redis.hset(self.get_redis_basekey()+":uid2flags",uid,flags)
        else:
            # unknown id
            pass

    def getMessage(self, uid):
        data = self.store.connection.msgget(self.guid, uid)
        if len(data) == 2:
            if (len(data[0]) >1):
                if (data[0][0][0] == '1'):
                    # success
                    result, uid, savedatestr, min_modseq = data[0][0].split('\t', 3)
                    message = Message()
                    # join without the last terminating dot line, include empty line for last CRLF
                    message.rfc822 = CRLF.join(data[0][1:-1])
                    message.received_date = int(savedatestr)
                    flags = self.store.redis.hget(self.get_redis_basekey()+":uid2flags",uid)
                    message.flag_new = 1 
                    self.getMessageFlagsFromFlags(flags, message)
                    return message 
        return None                  
                    
    def getMessageFlagsFromFlags(self, flags, message=None):
        if message == None:
            message = Message()
        extra_pos = -1
        
        flag = "\\Answered"
        pos = flags.find(flag)
        if (pos > -1):
            extra_pos = pos + len(flag)
            message.flag_answered = 1
        else:
            message.flag_answered = 0

        flag = "\\Flagged"
        pos = flags.find(flag)
        if (pos > -1):
            extra_pos = pos + len(flag)
            message.flag_flagged = 1
        else:
            message.flag_flagged = 0

        flag = "\\Seen"
        pos = flags.find(flag)
        if (pos > -1):
            extra_pos = pos + len(flag)
            message.flag_read = 1
        else:
            message.flag_read = 0

        flag = "\\Draft"
        pos = flags.find(flag)
        if (pos > -1):
            extra_pos = pos + len(flag)
            message.flag_draft = 1
        else:
            message.flag_draft = 0

        if extra_pos < 0 and  len(flags) > 0:
            extra_pos = 0
        if extra_pos > -1 and extra_pos < len(flags):
            message.extra_flags = flags[extra_pos:].strip()
        
        if (flags.find("$Forwarded") > -1):
            message.flag_forwarded = 1
        else:
            message.flag_forwarded = 0
        
        return message
        
    def getFlagsFromMessage(self, message):
        flag_list = []
        if (message.flag_answered): flag_list.append("\\Answered")
        if (message.flag_flagged): flag_list.append("\\Flagged")
        if (message.flag_read): flag_list.append("\\Seen")
        if (message.flag_draft): flag_list.append("\\Draft")
        if (message.flag_forwarded):
            if (message.extra_flags == None or message.extra_flags.find("$Forwarded") < 0): 
                flag_list.append("$Forwarded")
        if (message.extra_flags != None):
            flag_list.append(message.extra_flags)
        return " ".join(flag_list)

    def updateMessageReadFlag(self, id, flag_read):
        self.logger.debug("updateMessageReadFlag %s %s" % (id, flag_read))
        if not self.selected:
            self.store.connection.boxselect(self.guid)
            self.selected = True
        uid = self.getUidForGuid(id)
        if uid:
            message = Message()
            last_flags = self.store.redis.hget(self.get_redis_basekey()+":uid2flags",uid)
            if last_flags != None:
                self.getMessageFlagsFromFlags(last_flags, message)
                self.logger.debug("updateMessageReadFlag %s %s %s" % (id, flag_read, last_flags))
            message.flag_read = flag_read
            new_flags = self.getFlagsFromMessage(message)
            self.store.connection.msgupdate(uid, 0, new_flags)
            self.store.redis.hset(self.get_redis_basekey()+":uid2flags",uid,new_flags)
            self.logger.debug("updateMessageReadFlag %s %s uid %s %s" % (id, flag_read, uid, new_flags))
        else:
            self.logger.debug("updateMessageReadFlag %s %s UNKNOWN UID" % (id, flag_read))
        
    def delMessages(self, ids):
        returncode = 1
        if not self.selected:
            self.store.connection.boxselect(self.guid)
            self.selected = True
        del_uids = []
        for id in ids:
            uid = self.getUidForGuid(id)
            if uid:
                del_uids.append(uid)
            else:
                self.logger.debug("DEL NOT FOUND %s" % id)
                returncode = 0
        for del_uid in del_uids:
            self.store.connection.msgexpunge(del_uid)
        return returncode

    def begin(self):
        if self.store.redis.llen(self.get_redis_basekey()+":transactionlog") != 0:
            self.store.redis.delete(self.get_redis_basekey()+":transactionlog")
        
    def commit(self):
        if self.selected:
            alternative_folder_id = self.store.getAlternativeFolderId(self.guid)
            self.store.connection.boxselect(alternative_folder_id)
            self.selected = None
            # loop through the info feedback for uids
            uids = []
            while 1:
                if not self.store.connection.hasinfo():
                    if self.store.redis.llen(self.get_redis_basekey()+":transactionlog") == 0:
                        # not expecting any more new messages
                        break
                    else:
                        pass
                info = self.store.connection.getinfo()
                if (info[2]=='save'):
                    uid = info[3]['uid']
                    uids.append(uid)
                    # get id from transaction log
                    sequence_id = self.store.redis.lpop(self.get_redis_basekey()+":transactionlog")
                    id = self.store.redis.hget(self.get_redis_basekey()+":sequenceid2id",sequence_id)
                    # save uid to sourceid mapping
                    self.store.redis.hset(self.get_redis_basekey()+":uid2sourceid",uid,id)
                    # save sourceid to uid mapping
                    self.store.redis.hset(self.get_redis_basekey()+":sourceid2uid",id,uid)
            self.store.resync()
            new_max_modseq = self.store.get_folder_max_modseq(self.guid)
            expected_modseq = self.last_modseq + 1
            if (new_max_modseq == expected_modseq):
                # if yes then no concurrent imap changes, save uid modseq and uid flags
                for uid in uids:
                    self.store.redis.hset(self.get_redis_basekey()+":uid2modseq",uid,new_max_modseq)
            else:
#                # rescan mails to determine current modseq
                sequence_id = self.first_sequence_id
                for uid in uids:
                    self.store.redis.hset(self.get_redis_basekey()+":uid2expectedmodseq",uid,expected_modseq)
                    sequence_id = sequence_id + 1 
            # save the flag state we wrote
            sequence_id = self.first_sequence_id
            for uid in uids:
                flags = self.store.redis.hget(self.get_redis_basekey()+":sequenceid2flags",sequence_id)
                self.store.redis.hset(self.get_redis_basekey()+":uid2flags",uid,flags)
                sequence_id = sequence_id + 1 
            
            # save the last state   
            last_seen_modseq_str = self.store.redis.get(self.get_redis_basekey()+":last_seen_modseq")
            if (last_seen_modseq_str != None):
                last_seen_modseq = int(last_seen_modseq_str)
                if (new_max_modseq == last_seen_modseq + 1):
                    self.store.redis.set(self.get_redis_basekey()+":last_seen_modseq",new_max_modseq)
        else:
            # nothing to commit
            return
                    
    def getVersion(self):
        pass
    
    def updateChanges(self):
        self._loadChanges()
        changes = self.Changes
        connection = self.connection
        guid = self.guid
        
        last_known_change = changes.last()
        last_known_change_modseq = last_known_change.Id
        
        msg_list_data = connection.msglist(guid)
        
        for msg_list_item_line in msg_list_data:
            box_index, msg_guid, uid, modseq, flags, save_date = msg_list_item_line.split('\t')
    
    def getChanges(self, fromVersion):
        last_seen_modseq_str = self.store.redis.get(self.get_redis_basekey()+":last_seen_modseq")
        if (last_seen_modseq_str != None):
            last_seen_modseq = int(last_seen_modseq_str)
        else:
            last_seen_modseq = 0
        last_seen_deleted = 0
        first_dsync_expunged_uid = None
        last_seen_dsync_expunged_uid = None
        new_last_seen_dsync_expunged_uid = None
        known_dsync_expunged_uids = []
        # open folder
        # get msg-list
        min_modseq = last_seen_modseq + 1
        max_seen_modseq = last_seen_modseq
        msg_list_data = self.getMessageListForMinModseq(0)  # get all message infos
        next_uid = 1
        for msg_list_item in msg_list_data:
            uid = int(msg_list_item[2])
            flags = msg_list_item[4]
            if (uid != next_uid):
                # here's a gap check if we need to hard delete expunges
                deleted_uid = next_uid
                known_deleted_uids = []
                while deleted_uid < uid:
                    # check if there is a known gap up to before uid
                    last_seen_deleted_str = self.store.redis.hget(self.get_redis_basekey()+":uid2lastdeleted",deleted_uid)
                    if (last_seen_deleted_str != None):
                        # this deleted_uid is known to be deleted up until last_seen_deleted
                        last_seen_deleted = int(last_seen_deleted_str)
                        known_deleted_uids.append(deleted_uid)
                        deleted_uid = last_seen_deleted + 1
                    else:
                        # this deleted_uid is unknown
                        # create hard delete changes
                        deleted_id = self.store.redis.hget(self.get_redis_basekey()+":uid2sourceid",deleted_uid)
                        self.logger.debug("HARD DELETED UID: %d ID: %s" % (deleted_uid, deleted_id))
                        ics_key = str(min_modseq) + ":" + str(deleted_uid)
                        self.store.redis.hset(self.get_redis_basekey()+":changes",ics_key,"4") 
                        self.store.redis.hset(self.get_redis_basekey()+":tombstones",deleted_uid,min_modseq) 
                        deleted_uid = deleted_uid + 1
                # save known deleted gap
                self.store.redis.hset(self.get_redis_basekey()+":uid2lastdeleted",next_uid,uid-1) 
                # remove enclosed known gaps 
                for known_deleted_uid in known_deleted_uids:
                    if known_deleted_uid != next_uid:
                        self.store.redis.hdel(self.get_redis_basekey()+":uid2lastdeleted",known_deleted_uid)
                # adjust next_uid
                next_uid = uid
            # continue processing item
            next_uid = next_uid + 1
            # check modseq
            msg_modseq = int(msg_list_item[3])
            if msg_modseq > 0 and msg_modseq < min_modseq:
                # skip already processed changes
                continue
            if max_seen_modseq < msg_modseq:
                max_seen_modseq = msg_modseq
            if msg_modseq > 0:
                ics_key = str(msg_modseq) + ":" + str(uid)
            else:
                ics_key = str(min_modseq) + ":" + str(uid)
                self.logger.debug("ICSKEY: %s" % (ics_key))
            # is this uid known
            id = self.store.redis.hget(self.get_redis_basekey()+":uid2sourceid",uid)
            if (id == None):
                # unknown so it must be a imap side create
                self.logger.debug("CREATE NEW UID: %d" % uid)
                self.store.redis.hset(self.get_redis_basekey()+":uid2flags",uid,flags)
                self.store.redis.hset(self.get_redis_basekey()+":uid2modseq",uid,msg_modseq)
                self.store.redis.hset(self.get_redis_basekey()+":changes",ics_key,"1") 
                self.store.redis.rpush(self.get_redis_basekey()+":changeidqueue",ics_key) 
            else:
                # check modseq
                msg_last_seen_modseq_str = self.store.redis.hget(self.get_redis_basekey()+":uid2modseq",uid)
                if (msg_last_seen_modseq_str != None):
                    msg_last_seen_modseq = int(msg_last_seen_modseq_str)
                    if (msg_last_seen_modseq == msg_modseq):
                        # msg is unchanged
                        continue
                    else:
                        # msg has changed determine the changed flags
                        msg_last_seen_flags = self.store.redis.hget(self.get_redis_basekey()+":uid2flags",uid)
                        if (flags.find("\\Deleted") > -1):
                            self.logger.debug("MSG SOFT DELETED UID: %d" % uid)
                            self.store.redis.hset(self.get_redis_basekey()+":changes",ics_key,"3") 
                            self.store.redis.rpush(self.get_redis_basekey()+":changeidqueue",ics_key) 
                            self.store.redis.hset(self.get_redis_basekey()+":uid2flags",uid,flags)
                        elif (flags.find("\\dsync-expunged") > -1):
                            if (last_seen_dsync_expunged_uid != None and uid <= last_seen_dsync_expunged_uid):
                                # inside a block
                                if (new_last_seen_dsync_expunged_uid != None):
                                    # if new block is being expanden move along the new end
                                    new_last_seen_dsync_expunged_uid = uid
                                continue
                            if (last_seen_dsync_expunged_uid == None and first_dsync_expunged_uid == None):
                                # this is the beginning of the dsync-expunged block
                                first_dsync_expunged_uid = uid
                            last_seen_dsync_expunged_uid_str = self.store.redis.hget(self.get_redis_basekey()+":uid2lastdeleted",uid)
                            if (last_seen_dsync_expunged_uid_str != None):
                                last_seen_dsync_expunged_uid = int(last_seen_dsync_expunged_uid_str)
                                if (uid != first_dsync_expunged_uid):
                                    # remove embedded blocks afterwards
                                    known_dsync_expunged_uids.append(uid)
                                continue
                            self.logger.debug("MSG HARD DELETED UID: %d" % uid)
                            self.store.redis.hset(self.get_redis_basekey()+":changes",ics_key,"4") 
                            self.store.redis.rpush(self.get_redis_basekey()+":changeidqueue",ics_key) 
                            self.store.redis.hset(self.get_redis_basekey()+":tombstones",uid,min_modseq)
                            new_last_seen_dsync_expunged_uid = uid
                        else:
                            self.logger.debug("FLAGS CHANGED UID: %d FLAGS: %s LASTSEENFLAGS: %s" % (uid, flags, msg_last_seen_flags))
                            self.store.redis.hset(self.get_redis_basekey()+":changes",ics_key,"2:%s:%s" % (flags, msg_last_seen_flags)) 
                            self.store.redis.rpush(self.get_redis_basekey()+":changeidqueue",ics_key) 
                            self.store.redis.hset(self.get_redis_basekey()+":uid2flags",uid,flags)
                            self.store.redis.hset(self.get_redis_basekey()+":uid2modseq",uid,msg_modseq)
                else:
                    # get what was expected
                    msg_last_expected_modseq_str = self.store.redis.hget(self.get_redis_basekey()+":uid2expectedmodseq",uid)
                    if (msg_last_expected_modseq_str != None):
                        msg_last_expected_modseq = int(msg_last_expected_modseq_str)
                        if (msg_last_expected_modseq == msg_modseq):
                            # msg is unchanged and has the expected modseq
                            self.store.redis.hset(self.get_redis_basekey()+":uid2modseq",uid,msg_modseq)
                            self.store.redis.hdel(self.get_redis_basekey()+":uid2expectedmodseq",uid)
                        else:
                            # msg might be changed
                            msg_last_seen_flags = self.store.redis.hget(self.get_redis_basekey()+":uid2flags",uid)
                            if (msg_last_seen_flags == flags):
                                # unchanged this must be the assigned modseq
                                self.store.redis.hset(self.get_redis_basekey()+":uid2modseq",uid,msg_modseq)
                                self.store.redis.hdel(self.get_redis_basekey()+":uid2expectedmodseq",uid)
                            else:
                                # changed after save/mod
                                self.store.redis.hset(self.get_redis_basekey()+":uid2modseq",uid,msg_modseq)
                                self.store.redis.hdel(self.get_redis_basekey()+":uid2expectedmodseq",uid)
                                self.logger.debug("FLAGS CHANGED UID: %d FLAGS: %s MODSEQ %d" % (uid, flags, msg_modseq))
                                self.store.redis.hset(self.get_redis_basekey()+":uid2flags",uid,flags)
                                self.store.redis.hset(self.get_redis_basekey()+":changes",ics_key,"2:%s:%s" % (flags, msg_last_seen_flags))
                                self.store.redis.rpush(self.get_redis_basekey()+":changeidqueue",ics_key) 
                        
        if (first_dsync_expunged_uid != None):
            # check if the block has been enlarged
            if (new_last_seen_dsync_expunged_uid != None):
                self.store.redis.hset(self.get_redis_basekey()+":uid2lastdeleted",first_dsync_expunged_uid,new_last_seen_dsync_expunged_uid) 
            # remove enclosed known gaps of dsync-expunged
            for known_dsync_expunged_uid in known_dsync_expunged_uids:
                self.store.redis.hdel(self.get_redis_basekey()+":uid2lastdeleted",known_dsync_expunged_uid)
        self.store.redis.set(self.get_redis_basekey()+":last_seen_modseq", max_seen_modseq)                
    
    def _loadChanges(self):
        if self.Changes == None:
            changes = DsyncContentsChanges()
            changes.load()
            self.Changes = changes
