# -*- coding: utf-8; indent-tabs-mode: nil -*-

# mailbridge: a mail storage spanning bridge
# 
# Copyright 2011 Ryan Hughes <ryan@n-h.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import binascii

from MAPI.Util import *
from MAPI.Time import *
from MAPI.Struct import *
import inetmapi
import logging


class ZarafaStoreContentsImporter(object):
    '''
    classdocs
    '''


    def __init__(self, folder):
        '''
        Constructor
        '''
        self.folder = folder
        self.folder_source_key = self.folder.GetProps([PR_SOURCE_KEY], 0)[0].Value
        self.logger = logging.getLogger("ZarafaStoreContentsImporter")
        
    def Config(self, stream, flags):
        self.importer.Config(stream, flags)
    
    def UpdateState(self, stream):
        self.importer.UpdateState(stream)
        
    def importMessageChange(self, id, message):
        if (message.flag_new):
            msg = self.importer.ImportMessageChange([], 0)
            dopts = inetmapi.delivery_options()
            dopts.use_received_date = True
            if message.flag_read:
                dopts.mark_as_read = True
            else:
                dopts.mark_as_read = False
            dopts.user_entryid = None
            inetmapi.IMToMAPI(self.session, self.store, self.addressbook, msg, message.rfc822, dopts)
            msg.SaveChanges(KEEP_OPEN_READONLY)
            sourcekey = msg.GetProps([PR_SOURCE_KEY], 0)[0].Value
            return binascii.hexlify(sourcekey)
        else:
            msg = self.importer.ImportMessageChange([SPropValue(PR_SOURCE_KEY, binascii.unhexlify(id))], 0)
            flag_status_prop = msg.GetProps([PR_FLAG_STATUS], 0)[0]
            flag_icon_prop = msg.GetProps([PROP_TAG(PT_LONG, 0x1095)], 0)[0]
            icon_index_prop = msg.GetProps([PR_ICON_INDEX], 0)[0]
            props = []
            if message.flag_answered:
                props.append(SPropValue(PR_ICON_INDEX,261))
            elif message.flag_forwarded:
                props.append(SPropValue(PR_ICON_INDEX,262))
            else:
                props.append(SPropValue(PR_ICON_INDEX,0))
            if message.flag_flagged:
                if (PROP_TYPE(flag_icon_prop.ulPropTag)!=MAPICore.PT_ERROR and flag_icon_prop.Value > 0):
                    pass
                elif (PROP_TYPE(flag_status_prop.ulPropTag)!=MAPICore.PT_ERROR and flag_status_prop.Value == 2):
                    props.append(SPropValue(PROP_TAG(PT_LONG, 0x1095),6))
                else:
                    props.append(SPropValue(PROP_TAG(PT_LONG, 0x1095),6))
            if props:
                msg.SetProps(props)
            msg.SaveChanges(0)
            return None
        
    def importMessageDeletion(self, flags, ids):
        deleteids = []
        for id in ids:
            deleteids.append(binascii.unhexlify(id))
        return self.importer.ImportMessageDeletion(flags, deleteids)
        
    def importMessageReadStateChange(self, states):
        mapistates = []
        for state in states:
            mapiflags = 0L
            if (state.read_flag):
                mapiflags = mapiflags | MSGFLAG_READ
            else:
                pass
            mapistates.append(READSTATE(binascii.unhexlify(state.id), mapiflags))
            msg = self.importer.ImportMessageChange(
                        [SPropValue(PR_PARENT_SOURCE_KEY, self.folder_source_key),
                         SPropValue(PR_SOURCE_KEY, binascii.unhexlify(state.id))], 0)
            if (state.read_flag):
                msg.SetReadFlag(0)
            else:
                msg.SetReadFlag(CLEAR_READ_FLAG)
            msg.SaveChanges(0)
        pass
        

class ZarafaStoreHierarchyImporter(object):
    '''
    classdocs
    '''


    def __init__(self, folder):
        '''
        Constructor
        '''
        self.folder = folder
        self.folder_source_key = self.folder.GetProps([PR_SOURCE_KEY], 0)[0].Value
        
    def Config(self, stream, flags):
        self.importer.Config(stream, flags)
    
    def UpdateState(self, stream):
        self.importer.UpdateState(stream)
        
    def importFolderChange(self, id, folder):
        if (id == None):
            # create new folder in mapi
            # open parent folder
            ema = self.store.QueryInterface(IID_IExchangeManageStore)
            if (len(folder.parent) > 0):
                parent_source_key = binascii.unhexlify(folder.parent)
            else:
                parent_source_key = self.backend.ipm_root_sourcekey
            parent_entryid = ema.EntryIDFromSourceKey(parent_source_key, '')
            parent_mapifolder = self.store.OpenEntry(parent_entryid, None, MAPI_MODIFY)
            mapifolder = parent_mapifolder.CreateFolder(FOLDER_GENERIC, folder.name, u'', None, MAPI_UNICODE)
            sourcekey = mapifolder.GetProps([PR_SOURCE_KEY], 0)[0].Value
            self.importer.ImportFolderChange([SPropValue(PR_PARENT_SOURCE_KEY, parent_source_key),
                                              SPropValue(PR_SOURCE_KEY, sourcekey),
                                              SPropValue(PR_DISPLAY_NAME_W, folder.name)])
            return binascii.hexlify(sourcekey)
        else:
            if (len(folder.parent) > 0):
                parent_source_key = binascii.unhexlify(folder.parent)
            else:
                parent_source_key = self.backend.ipm_root_sourcekey
            sourcekey = binascii.unhexlify(folder.id)
            if (self.backend.special_source_keys.has_key(sourcekey)):
                # don't modify special folders
                return None
            self.importer.ImportFolderChange([SPropValue(PR_PARENT_SOURCE_KEY, parent_source_key),
                                              SPropValue(PR_SOURCE_KEY, sourcekey),
                                              SPropValue(PR_DISPLAY_NAME_W, folder.name)])
            return None
        
    def importFolderDeletion(self, flags, ids):
        deleteids = []
        for id in ids:
            deleteids.append(binascii.unhexlify(id))
        return self.importer.ImportFolderDeletion(flags, deleteids)
        
